import abc
from aoa.api.base_api import BaseApi
from aoa.api.api_iterator import ApiIterator


class IteratorBaseApi(BaseApi):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def find_all(self):
        pass

    def __iter__(self):
        return ApiIterator(api_func=self.find_all, entities=self.path.split('/')[-2])

    def __len__(self):
        return self.find_all()['page']['totalElements']



import base64, secrets

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

backend = default_backend()


default_conf = {
    "iterations": 65536,
    "bytes_key_length": 32,
    "bytes_salt_length": 16,
    "bytes_iv_length": 16,
    "bytes_auth_tag_length": 16
}


def _derive_key(password, salt, iterations, length):
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(), length=length, salt=salt,
        iterations=iterations, backend=backend)
    return kdf.derive(password)


def aes_gcm_encrypt(message: str, password: str, conf: dict = default_conf) -> str:
    """

    :param message: the text to encrypt
    :param password: the token/key to use for decryption (no spaces allowed)
    :param conf: if not specified will use the default conf of
            default_conf = {
                "iterations": 65536,
                "bytes_key_length": 32,
                "bytes_salt_length": 16,
                "bytes_iv_length": 16,
                "bytes_auth_tag_length": 16
            }
    :return: the encrypted message
    """
    auth_tag = secrets.token_bytes(conf["bytes_auth_tag_length"])
    iv = secrets.token_bytes(conf["bytes_iv_length"])
    salt = secrets.token_bytes(conf["bytes_salt_length"])
    key = _derive_key(password.encode(), salt, conf["iterations"], conf["bytes_key_length"])
    algorithm = algorithms.AES(key)
    cipher = Cipher(algorithm, modes.GCM(iv), backend=backend)
    encryptor = cipher.encryptor()
    ciphertext = encryptor.update(message.encode())

    return base64.urlsafe_b64encode(iv + salt + ciphertext + auth_tag).decode('utf-8')


def aes_gcm_decrypt(message: str, password: str, conf: dict = default_conf) -> str:
    """
    Decrypt text which was previously encrypted by the AOA core for e.g connection credentials. It is AES 256 bit HMAC

    :param message: the text to decrypt
    :param password: the token/key to use for decryption
    :param conf: if not specified will use the default conf of
            default_conf = {
                "iterations": 65536,
                "bytes_key_length": 32,
                "bytes_salt_length": 16,
                "bytes_iv_length": 16,
                "bytes_auth_tag_length": 16
            }
    :return: decrypted text
    """
    data = base64.urlsafe_b64decode(message.encode())
    (iv, salt, ciphertext, auth_tag) = data[:conf["bytes_iv_length"]], \
                                       data[conf["bytes_iv_length"] : conf["bytes_iv_length"] + conf["bytes_salt_length"]], \
                                       data[conf["bytes_iv_length"] + conf["bytes_salt_length"] : -conf["bytes_auth_tag_length"]], \
                                       data[-conf["bytes_auth_tag_length"]:]
    key = _derive_key(password.encode(), salt, iterations=conf["iterations"], length=conf["bytes_key_length"])
    algorithm = algorithms.AES(key)
    cipher = Cipher(algorithm, modes.GCM(iv, salt), backend=backend)
    decryptor = cipher.decryptor()
    decrypted = decryptor.update(ciphertext)

    return decrypted.decode('utf-8')



from __future__ import absolute_import
from typing import List
from aoa.api.base_api import BaseApi

import requests
import uuid
import os


class TrainedModelArtefactsApi(BaseApi):

    def list_artefacts(self, trained_model_id: uuid):
        """
        returns all trained models

        Parameters:
           trained_model_id (uuid): Trained Model Id

        Returns:
            (list): all trained model artefacts
        """
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept(['application/json'])]
        header_params = self.generate_params(header_vars, header_vals)

        return self.aoa_client.get_request(
            "/api/trainedModels/{}/artefacts/listObjects".format(trained_model_id),
            header_params,
            query_params=None)["objects"]

    def get_signed_download_url(self, trained_model_id: uuid, artefact: str):
        """
        returns a signed url for the artefact

        Parameters:
           trained_model_id (uuid): Trained Model Id
           artefact (str): The artefact to generate the signed url for

        Returns:
            (str): the signed url
        """
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept(['application/json'])]
        header_params = self.generate_params(header_vars, header_vals)
        query_params = self.generate_params(['objectKey'], [artefact])

        response = self.aoa_client.get_request(
            "/api/trainedModels/{}/artefacts/signedDownloadUrl".format(trained_model_id),
            header_params,
            query_params)

        return response["endpoint"]

    def get_signed_upload_url(self, trained_model_id: uuid, artefact: str):
        """
        returns a signed url for the artefact

        Parameters:
           trained_model_id (uuid): Trained Model Id
           artefact (str): The artefact to generate the signed url for

        Returns:
            (str): the signed url
        """
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept(['application/json'])]
        header_params = self.generate_params(header_vars, header_vals)
        query_params = self.generate_params(['objectKey'], [artefact])

        response = self.aoa_client.get_request(
            "/api/trainedModels/{}/artefacts/signedUploadUrl".format(trained_model_id),
            header_params,
            query_params)

        return response["url"]

    def download_artefacts(self, trained_model_id: uuid, path: str = "."):
        """
        downloads all artefacts for the given trained model

        Parameters:
           trained_model_id (uuid): Trained Model Id
           path (str): the path to download the artefacts to (default cwd)

        Returns:
            None
        """

        for artefact in self.list_artefacts(trained_model_id):
            response = self.aoa_client.session.get(self.get_signed_download_url(trained_model_id, artefact))

            output_file = "{}/{}".format(path, artefact)
            if not os.path.exists(os.path.dirname(output_file)):
                os.makedirs(os.path.dirname(output_file))

            with open(output_file, "wb") as f:
                for chunk in response.iter_content(chunk_size=1024*1024):
                    f.write(chunk)

    def upload_artefacts(self, trained_model_id: uuid, artefacts: List = None, artefacts_folder: str = None):
        """
        uploads artefacts for the given trained model

        Parameters:
           trained_model_id (uuid): Trained Model Id
           artefacts (List): The artefact paths to upload (must specify this or artefacts_folder)
           artefacts_folder (str): The artefact folder (must specify this or artefacts list)
        Returns:
            None
        """

        if artefacts is None and artefacts_folder is None:
            raise ValueError("Either artefacts or artefacts_folder argument must be specified")

        if artefacts is not None:
            for artefact in artefacts:
                object_key = os.path.basename(artefact)
                self.__upload_artefact(artefact, object_key, trained_model_id)

        else:
            for root, d, files in os.walk(artefacts_folder):
                for file in files:
                    object_key = os.path.relpath(os.path.join(root, file), artefacts_folder)
                    self.__upload_artefact(os.path.join(root, file), object_key, trained_model_id)

    def __upload_artefact(self, artefact, object_key, trained_model_id):
        query_params = {
            'objectKey': object_key
        }
        header_params = {
            'AOA-Project-ID': "{}".format(self.aoa_client.project_id)
        }
        signed_url = self.aoa_client.get_request("/api/trainedModels/{}/artefacts/signedUploadUrl"
                                                 .format(trained_model_id), header_params, query_params)
        upload_resp = requests.put(signed_url['endpoint'], data=open(artefact, 'rb'))
        upload_resp.raise_for_status()



from requests.auth import HTTPBasicAuth

import requests
import json


class AutomationClient():

    def __init__(self, url, auth_mode="basic", auth_user=None, auth_pass=None):
        self.url = url

        if auth_mode == "basic":
            self.auth = HTTPBasicAuth(auth_user, auth_pass)

        elif auth_mode == "kerberos":
            from requests_kerberos import HTTPKerberosAuth
            self.auth=HTTPKerberosAuth()

        else:
            raise Exception("Auth mode: {} not supported.".format(auth_mode))

    def trigger_workflow(self, event):
        resp = requests.post(
            url=self.url,
            auth=self.auth,
            headers={'Content-Type': 'application/json'},
            data=json.dumps(event)
        )

        resp.raise_for_status()

        try:
            return resp.json()
        except ValueError:
            return resp.text



from __future__ import absolute_import

import json
import os
import yaml
import logging
import requests
from requests.auth import AuthBase
from typing import List, Dict


class AoaClient(object):
    def __init__(self, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.project_id = None
        self.aoa_url = None
        self.s3_bucket = None

        self.session = requests.Session()
        self.__parse_aoa_config(**kwargs)

    def __parse_yaml(self, yaml_path: str):
        with open(yaml_path, "r") as handle:
            conf = yaml.safe_load(handle)
        self.__parse_kwargs(**conf)

    def __parse_kwargs(self, **kwargs):
        self.aoa_url = kwargs.get("aoa_url", self.aoa_url)

        if "auth_mode" in kwargs:
            auth_mode = kwargs["auth_mode"]
            if auth_mode == "basic":
                credentials = kwargs.get("auth_credentials", None)
                # backward compatibility for credentials key as was called aoa_credentials instead of auth_credentials
                if "aoa_credentials" in kwargs and credentials is None:
                    credentials = kwargs.get("aoa_credentials")

                self.__configure_basic_auth(credentials,
                                            kwargs.get("auth_user", None),
                                            kwargs.get("auth_pass", None))

            elif auth_mode == "kerberos":
                self.__configure_kerberos()

            elif auth_mode == "oauth":
                self.__configure_oauth_refresh_token(
                    kwargs.get("auth_client_id", None),
                    kwargs.get("auth_client_secret", None),
                    kwargs.get("auth_client_token_url", None),
                    kwargs.get("auth_client_refresh_token", None))

            else:
                raise Exception("Auth mode: {} not supported.".format(auth_mode))

        if "verify_connection" in kwargs:
            self.verify_aoa_connection = kwargs["verify_connection"]

    def __parse_env_variables(self):
        self.aoa_url = os.environ.get("AOA_URL", self.aoa_url)

        if "AOA_API_AUTH_MODE" in os.environ:
            auth_mode = os.environ.get("AOA_API_AUTH_MODE")

            if auth_mode == "basic":
                self.__configure_basic_auth(os.environ.get("AOA_API_AUTH_CREDENTIALS", None),
                                            os.environ.get("AOA_API_AUTH_USER", None),
                                            os.environ.get("AOA_API_AUTH_PASS", None))

            elif auth_mode == "kerberos":
                self.__configure_kerberos()

            elif auth_mode == "oauth":
                self.__configure_oauth_refresh_token(
                    os.environ.get("AOA_API_AUTH_CLIENT_ID", None),
                    os.environ.get("AOA_API_AUTH_CLIENT_SECRET", None),
                    os.environ.get("AOA_API_AUTH_TOKEN_URL", None),
                    os.environ.get("AOA_API_AUTH_REFRESH_TOKEN", None))

            else:
                raise Exception("Auth mode: {} not supported.".format(auth_mode))

    def __parse_aoa_config(self, **kwargs):
        if "config_file" in kwargs:
            self.__parse_yaml(kwargs['config_file'])
        else:
            from pathlib import Path
            config_file = "{}/.aoa/config.yaml".format(Path.home())
            if os.path.isfile(config_file):
                self.__parse_yaml(config_file)

        self.__parse_env_variables()
        self.__parse_kwargs(**kwargs)

    def set_project_id(self, project_id: str):
        """
        set project id

        Parameters:
           project_id (str): project id(uuid)
        """
        self.project_id = project_id

    def get_current_project(self):
        """
        get project id

        Return:
           project_id (str): project id(uuid)
        """
        return self.project_id

    def select_header_accept(self, accepts: List[str]):
        """
        converts list of header into a string

        Return:
            (str): request header
        """
        if not accepts:
            return
        accepts = [x.lower() for x in accepts]
        if 'application/json' in accepts:
            return 'application/json'
        else:
            return ', '.join(accepts)

    def get_request(self, path, header_params: Dict[str, str], query_params: Dict[str, str]):
        """
        wrapper for get request

        Parameters:
           path (str): url
           header_params (dict): header parameters
           query_params (dict): query parameters

        Returns:
            dict for resources, str for errors, None for 404
        Raise:
            raises HTTPError in case of error status code other than 404
        """

        resp = self.session.get(
            url=self.__strip_url(self.aoa_url) + path,
            headers=header_params,
            params=query_params
        )

        if resp.status_code == 404:
            return None

        resp.raise_for_status()

        try:
            return resp.json()
        except ValueError:
            return resp.text

    def post_request(self, path, header_params: Dict[str, str], query_params: Dict[str, str], body: Dict[str, str]):
        """
        wrapper for post request

        Parameters:
           path (str): url
           header_params (dict): header parameters
           query_params (dict): query parameters
           body (dict): request body

        Returns:
            dict for resources, str for errors
        Raise:
            raises HTTPError in case of error status code
        """
        resp = self.session.post(
            url=self.__strip_url(self.aoa_url) + path,
            headers=header_params,
            params=query_params,
            data=json.dumps(body)
        )

        resp.raise_for_status()

        try:
            return resp.json()
        except ValueError:
            return resp.text

    def put_request(self, path, header_params: Dict[str, str], query_params: Dict[str, str], body: Dict[str, str]):
        """
        wrapper for put request

        Parameters:
           path (str): url
           header_params (dict): header parameters
           query_params (dict): query parameters
           body (dict): request body

        Returns:
            dict for resources, str for errors
        Raise:
            raises HTTPError in case of error status code
        """
        resp = self.session.put(
            url=self.__strip_url(self.aoa_url) + path,
            headers=header_params,
            params=query_params,
            data=json.dumps(body)
        )

        resp.raise_for_status()

        try:
            return resp.json()
        except ValueError:
            return resp.text

    def __strip_url(self, url):
        return url.rstrip('/')

    def __configure_basic_auth(self, credentials, username, password):
        if credentials is None and (username is None and password is None):
            raise Exception("credentials or (username, password) must be defined with auth_mode of 'basic'")

        from base64 import b64encode

        if credentials:
            self.session.auth = AoaAuth(credentials)

        else:
            self.session.auth = AoaAuth(b64encode("{}:{}".format(username, password).encode()).decode())

    def __configure_oauth_refresh_token(self, client_id, client_secret, token_url, refresh_token):
        if client_id is None or client_secret is None or token_url is None or refresh_token is None:
            raise Exception("client_id, client_secret, token_url and refresh_token must be defined "
                            "with auth_mode of 'oauth'")

        from requests_oauthlib import OAuth2Session
        from requests.auth import HTTPBasicAuth

        self.session = OAuth2Session(client_id=client_id)
        self.session.refresh_token(token_url=token_url, refresh_token=refresh_token,
                                   auth=HTTPBasicAuth(client_id, client_secret))

    def __configure_kerberos(self):
        from requests_kerberos import HTTPKerberosAuth, OPTIONAL
        self.session.auth = HTTPKerberosAuth(mutual_authentication=OPTIONAL)


class AoaAuth(AuthBase):
    def __init__(self, aoa_credentials):
        self.aoa_credentials = aoa_credentials

    def __call__(self, r):
        r.headers['Authorization'] = "Basic {}".format(self.aoa_credentials)
        return r



from __future__ import absolute_import

from typing import List, Dict

from aoa.api_client import AoaClient


class BaseApi(object):
    path = ''

    def __init__(self, aoa_client=None):
        """
        constructor for api

        Parameters:
           aoa_client (AoaClient): AoaClient to use with api, if not specified will create a new AoaClient object
        """
        if aoa_client is None:
            aoa_client = AoaClient()
        self.aoa_client = aoa_client

    def generate_params(self, header_params: List[str], header_vals: List[str]):
        """
        returns generated parameters

        Parameters:
           header_params (List[str]): list of parameter names
           header_vals (List[str]): list of parameter values

        Returns:
            (dict): generated parameters
        """
        return dict(zip(header_params, header_vals))

    def required_params(self, param_names: List[str], dict_obj: Dict[str, str]):
        """
        checks required parameters, raises exception if the required parameter is missing in the dictionary

        Parameters:
           param_names (List[str]): list of required parameter names
           dict_obj (Dict[str, str]): dictionary to check for required parameters
        """
        for param in param_names:
            if param not in dict_obj:
                raise ValueError("Missing required value " + str(param))




import urllib.request


def read_dataset_from_url(spark, url):
    # for this demo we're downloading the dataset locally and then reading it. This is obviously not production setting
    # https://raw.githubusercontent.com/apache/spark/branch-2.4/data/mllib/sample_libsvm_data.txt
    urllib.request.urlretrieve(url, "/tmp/data.txt")
    return spark.read.format("libsvm").load("/tmp/data.txt")


from aoa.model_utility.base_model import BaseModel
import json
import os
import importlib
import logging
import shutil
import sys


class ScoreModel(BaseModel):

    def __init__(self, model_utility):
        super().__init__(model_utility)
        self.logger = logging.getLogger(__name__)

    def batch_score_model_local(self, model_id: str = None, base_path: str = None, data_conf: dict = None):
        base_path = self.model_utility.model_catalog_path if base_path is None else os.path.join(base_path, '')
        model_path = base_path + 'model_definitions/'

        if not os.path.exists(model_path):
            raise ValueError("model directory {0} does not exist".format(model_path))

        if model_id is None:
            if "id" in self.model_utility.model:
                model_id = self.model_utility.model["id"]
            else:
                BaseModel.get_model_id(model_path)

        model_artefacts_path = ".artefacts/{}".format(model_id)

        if not os.path.exists("{}/output".format(model_artefacts_path)):
            raise ValueError("You must run training before trying to run scoring.")

        try:
            self.__cleanup()

            os.makedirs("./artifacts")

            os.symlink("../{}/output/".format(model_artefacts_path), "./artifacts/input", target_is_directory=True)

            # legacy models folder
            if not os.path.exists("./models"):
                os.symlink("{}/output".format(model_artefacts_path), "./models", target_is_directory=True)

            model_dir = model_path + model_id

            with open(model_dir + "/model.json", 'r') as f:
                model_definition = json.load(f)

            with open(model_dir + "/config.json", 'r') as f:
                model_conf = json.load(f)

            self.logger.info("Loading and executing model code")

            cli_model_kargs = {
                "model_id": model_id,
                "model_version": "cli",
                "project_id": self._BaseModel__get_project_id()
            }

            engine = self._BaseModel__get_engine(model_definition, "deployment")
            if engine == "python" or engine == "pyspark":
                if engine == "pyspark":
                    self._BaseModel__configure_pyspark(model_definition, model_artefacts_path, model_dir, "deployment")

                sys.path.append(model_dir)
                scoring = importlib.import_module(".scoring", package="model_modules")
                scoring.score(data_conf, model_conf, **cli_model_kargs)

            elif engine == "sql":
                raise Exception("not supported")

            elif engine == "R":
                self._BaseModel__run_r_model(model_id, base_path, data_conf, "score.batch")

            else:
                raise Exception("Unsupported language: {}".format(model_definition["language"]))

            self.__cleanup()
        except:
            self.__cleanup()
            self.logger.exception("Exception running model code")
            raise

    def __cleanup(self):
        if os.path.exists("./artifacts"):
            shutil.rmtree("./artifacts")
        try:
            os.remove("./models")
        except FileNotFoundError:
            self.logger.debug("Nothing to remove, folder './models' does not exists.")



from __future__ import absolute_import
from typing import Dict

from aoa.api.iterator_base_api import IteratorBaseApi


class DatasetConnectionApi(IteratorBaseApi):
    path = "/api/datasetConnections/"

    def find_all(self, projection: str = None, page: int = None, size: int = None, sort: str = None):
        """
        returns all dataset connections of a project

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): all dataset connections
        """
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept([
                'application/json',
                'application/hal+json',
                'text/uri-list',
                'application/x-spring-data-compact+json'])]
        header_params = self.generate_params(header_vars, header_vals)

        query_vars = ['projection', 'sort', 'page', 'size', 'sort']
        query_vals = [projection, sort, page, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            self.path,
            header_params,
            query_params)

    def find_by_id(self, dataset_connection_id: str, projection: str = None):
        """
        returns a dataset connection of a project

        Parameters:
           dataset_connection_id (str): dataset connection id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): dataset connection
        """
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept([
                'application/json',
                'application/hal+json',
                'text/uri-list',
                'application/x-spring-data-compact+json'])]
        header_params = self.generate_params(header_vars, header_vals)

        query_vars = ['projection']
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            self.path + dataset_connection_id,
            header_params,
            query_params)

    def render_vars(self, dataset_connection_id: str):
        """
        returns dictionary with the env vars for a given dataset connection of a project

        Parameters:
           dataset_connection_id (str): dataset connection id(uuid) to find

        Returns:
            (dict): dataset connection envs
        """
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept([
                'application/json',
                'application/hal+json',
                'text/uri-list',
                'application/x-spring-data-compact+json'])]
        header_params = self.generate_params(header_vars, header_vals)

        return self.aoa_client.get_request(
            self.path + dataset_connection_id + "/renderVars",
            header_params,
            dict())

    def save(self, dataset_connection: Dict[str, str]):
        """
        register a dataset connection

        Parameters:
           dataset connection (dict): dataset connection to register

        Returns:
            (dict): dataset template
        """
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept([
                'application/json',
                'application/hal+json',
                'text/uri-list',
                'application/x-spring-data-compact+json'])]
        header_params = self.generate_params(header_vars, header_vals)

        self.required_params(['name', 'description', 'type', 'metadata', 'credentials'], dataset_connection)

        query_params = {}

        return self.aoa_client.post_request(
            self.path,
            header_params,
            query_params,
            dataset_connection)



from teradataml.dataframe.dataframe import DataFrame
from typing import List

import json


def save_evaluation_metrics(partition_df: DataFrame, metrics: List):
    """
    :param partition_df: teradata dataframe containing at least ["partition_id", "partition_metadata", "num_rows"]
    :param metrics: list of metrics to normalize and report
    :return: None
    """
    total_rows = int(partition_df.select(["num_rows"]).sum().to_pandas().iloc[0])

    metrics_sql = [f"SUM(CAST(partition_metadata AS JSON).JSONExtractValue('$.metrics.{metric}') * num_rows/{total_rows}) AS {metric}" for metric in metrics]
    joined_metrics_sql = ','.join(metrics_sql)
    metrics = DataFrame.from_query(f"SELECT {joined_metrics_sql} FROM {partition_df._table_name}").to_pandas()

    metrics = {metric: "{:.2f}".format(metrics.iloc[0][metric]) for metric in metrics}

    with open("artifacts/output/metrics.json", 'w+') as f:
        json.dump(metrics, f, indent=2)


def save_metadata(partition_df: DataFrame):
    """
    create statistic summaries based on the provided dataframe produced via training or evaluation

    partitions.json is {
        "<partition1 key>": <partition1_metadata>,
        "<partition2 key>": <partition2_metadata>,
        ...
    }

    data_stats.json is {
        "num_rows": <num_rows>,
        "num_partitions": <num_partitions>
    }

    :param partition_df: teradata dataframe containing at least ["partition_id", "partition_metadata", "num_rows"]
    :return: None
    """

    total_rows = int(partition_df.select(["num_rows"]).sum().to_pandas().iloc[0])

    metadata_df = partition_df.select(["partition_id", "partition_metadata", "num_rows"]).to_pandas()

    metadata_dict = {r["partition_id"]: json.loads(r["partition_metadata"])
                     for r in metadata_df.to_dict(orient='records')}

    with open("artifacts/output/partitions.json", 'w+') as f:
        json.dump(metadata_dict, f, indent=2)

    data_metadata = {
        "num_rows": total_rows,
        "num_partitions": int(metadata_df.shape[0])
    }

    with open("artifacts/output/data_stats.json", 'w+') as f:
        json.dump(data_metadata, f, indent=2)


def cleanup_cli(model_version: str):
    """
    cli uses model version of "cli" always. We need to cleanup models table between runs.
    A better solution would be for the cli to write to a different table completely and just "recreate" on each run

    :param model_version: the model version being executed
    :return: None
    """
    if model_version == "cli":
        from teradataml.context.context import get_connection
        get_connection().execute("DELETE FROM aoa_sto_models WHERE model_version='cli'")




def score(data_conf, model_conf, **kwargs):
    """Python score method called by AOA framework batch mode

    Parameters:
    data_conf (dict): The dataset metadata
    model_conf (dict): The model configuration to use

    Returns:
    None:No return

    """


def evaluate(data_conf, model_conf, **kwargs):
    """Python evaluate method called by AOA framework

    Parameters:
    data_conf (dict): The dataset metadata
    model_conf (dict): The model configuration to use

    Returns:
    None:No return

    """

    # dump results as json file evaluation.json to models/ folder
    print("Evaluation complete...")


# Uncomment this code if you want to deploy your model as a Web Service (Real-time / Interactive usage)
# class ModelScorer(object):
#    def __init__(self, config=None):
#        self.model = joblib.load('models/iris_knn.joblib')
#
#    def predict(self, data):
#        return self.model.predict([data])
#


import pandas as pd
from sklearn import metrics
from sklearn.externals import joblib
import json


def score(data_conf, model_conf, **kwargs):
    predict_df = pd.read_csv(data_conf['location'])
    features = 'sepallength,sepalwidth,petallength,petalwidth'.split(',')
    X_test = predict_df.loc[:, features]
    y_test = predict_df['class']

    model = joblib.load('artifacts/input/iris_knn.joblib')

    y_pred = model.predict(X_test)

    print("Finished Scoring")

    # store predictions somewhere.. As this is demo, we'll just print to stdout.
    print(y_pred)

    return X_test, y_pred, y_test, model


def save_plot(title):
    import matplotlib.pyplot as plt

    plt.title(title)
    fig = plt.gcf()
    filename = title.replace(" ", "_").lower()
    fig.savefig('artifacts/output/{}'.format(filename), dpi=500)


def evaluate(data_conf, model_conf, **kwargs):
    """Python evaluate method called by AOA framework

    Parameters:
    data_conf (dict): The dataset metadata
    model_conf (dict): The model configuration to use

    Returns:
    None:No return

    """

    X_test, y_pred, y_test, model = score(data_conf, model_conf, **kwargs)

    evaluation = {
        'Accuracy': '{:.2f}'.format(metrics.accuracy_score(y_test, y_pred)),
        'Recall': '{:.2f}'.format(metrics.recall_score(y_test, y_pred)),
        'Precision': '{:.2f}'.format(metrics.precision_score(y_test, y_pred)),
        'f1-score': '{:.2f}'.format(metrics.f1_score(y_test, y_pred))
    }

    with open("artifacts/output/metrics.json", "w+") as f:
        json.dump(evaluation, f)

    metrics.plot_confusion_matrix(model, X_test, y_test)
    save_plot('Confusion Matrix')

    metrics.plot_roc_curve(model, X_test, y_test)
    save_plot('ROC Curve')

    print("Evaluation complete...")


# Uncomment this code if you want to deploy your model as a Web Service (Real-time / Interactive usage)
# class ModelScorer(object):
#    def __init__(self, config=None):
#        self.model = joblib.load('artifacts/input/iris_knn.joblib')
#
#    def predict(self, data):
#        return self.model.predict([data])
#


from __future__ import absolute_import

class ApiIterator(object):
    _pos = 0
    _current_page = 0

    def __init__(self, api_func = None, entities: str = None,  projection: str = None, page_size: int = None, sort: str = None):
        if api_func is None or entities is None:
            pass
        else:
            self._api_func = api_func
            self._projection = projection
            self._page_size = page_size
            self._sort = sort
            page_info = api_func(projection = projection, page = 0, size = page_size, sort = sort)['page']
            self._total_elements = page_info['totalElements']
            self._total_pages = page_info['totalPages']
            self._cache = api_func(projection = projection, page = 0, size = page_size, sort = sort)['_embedded'][entities]
            self._entities = entities

    def __next__(self):
        if self._pos < len(self._cache):
            result = self._cache[self._pos]
            self._pos += 1
            return result
        elif self._current_page < self._total_pages - 1:
            self._current_page += 1
            self._pos = 1 # so we skip the first one in next iteration
            self._cache = self._api_func(projection =  self._projection, page = self._current_page, size = self._page_size, sort = self._sort)['_embedded'][self._entities]
            return self._cache[0]
        else:
            raise StopIteration




from setuptools import setup, find_packages
from aoa import __version__

NAME = "aoa"

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open("docs/pypi.md", "r") as fh:
    long_description = fh.read()

setup(
    name=NAME,
    version=__version__,
    description="Python client for Teradata AnalyticOps Accelerator (AOA)",
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.5.2',
    author="Teradata",
    author_email="teradata.corporation@teradatacorporation.com",
    url="",
    install_requires=required,
    extra_require={
        "kerberos": [
            "requests_kerberos>=0.12.0",
            "pykerberos==1.1.14"
        ]
    },
    tests_require=['pytest', 'pytest-console-scripts'],
    setup_requires=['pytest-runner'],
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    scripts=['scripts/aoa']
)



import json
import logging

from pyspark.ml.classification import LogisticRegressionModel
from pyspark.ml.evaluation import BinaryClassificationEvaluator
from pyspark import SparkConf
from pyspark.sql import SparkSession
from pyspark.sql.functions import monotonically_increasing_id
from .util import read_dataset_from_url

logging.getLogger("py4j").setLevel(logging.ERROR)

# spark name and other properties are set by the framework launcher in spark-submit. don't override
spark = SparkSession.builder \
    .config(conf=SparkConf()) \
    .getOrCreate()


def score(data_conf, model_conf, **kwargs):
    lr_model = LogisticRegressionModel.load(spark.conf.get("spark.aoa.modelPath"))

    test = read_dataset_from_url(spark, data_conf["url"])
    # we need to ensure that test has an id for comparison (almost all datasets will already have a primary key!!)
    test = test.select("*").withColumn("id", monotonically_increasing_id())

    predictions = lr_model.transform(test).select("id", "rawPrediction", "prediction", "probability")

    predictions.write.mode("overwrite").save(data_conf["predictions"])
    logging.info("Finished scoring")


def evaluate(data_conf, model_conf, **kwargs):

    score(data_conf, model_conf, **kwargs)

    test = read_dataset_from_url(spark, data_conf["url"])

    expected = test.select("*").withColumn("id", monotonically_increasing_id())
    actual = spark.read.load(data_conf["predictions"])

    evaluator = BinaryClassificationEvaluator()
    roc = evaluator.evaluate(expected.join(actual, expected.id == actual.id))
    logging.info('Test Area Under ROC: {}'.format(roc))

    with open("models/evaluation.json", "w+") as f:
        json.dump({'roc': roc}, f)

    logging.info("Finished Evaluation")



from __future__ import absolute_import
from typing import Dict

from aoa.api.base_api import BaseApi


class MessageApi(BaseApi):

    path = "/api/message"

    def send_job_event(self, event: Dict):
        """
        send a job event

        Parameters:
           event (dict): event to send

        Returns:
            (dict): event
        """

        return self._send_event("jobevents", event)

    def send_progress_event(self, event: Dict):
        """
        send a progress event

        Parameters:
           event (dict): event to send

        Returns:
            (dict): event
        """

        return self._send_event("jobprogress", event)

    def _send_event(self, topic, event):

        header_params = {'Content-Type': 'application/json'}
        query_params = {'type': 'topic'}

        return self.aoa_client.post_request(
            "{}/{}".format(self.path, topic),
            header_params,
            query_params,
            event)




import logging

from pyspark.ml.classification import LogisticRegression
from pyspark import SparkConf
from pyspark.sql import SparkSession
from .util import read_dataset_from_url

logging.getLogger("py4j").setLevel(logging.ERROR)

# spark name and other properties are set by the framework launcher in spark-submit. don't override
spark = SparkSession.builder \
    .config(conf=SparkConf()) \
    .getOrCreate()


def train(data_conf, model_conf, **kwargs):
    hyperparams = model_conf["hyperParameters"]

    train = read_dataset_from_url(spark, data_conf["url"])

    lr = LogisticRegression(maxIter=hyperparams["maxIter"],
                            regParam=hyperparams["regParam"],
                            elasticNetParam=hyperparams["elasticNetParam"])

    logging.info("Starting training...")

    lr_model = lr.fit(train)

    # Print the coefficients and intercept for logistic regression
    logging.debug("Coefficients: {}".format(str(lr_model.coefficients)))
    logging.debug("Intercept: {}".format(str(lr_model.intercept)))

    logging.info("Finished training")

    # export model artefacts to models/ folder
    lr_model.write().save(spark.conf.get("spark.aoa.modelPath"))
    logging.info("Saved trained model")

    return lr_model


import logging
import json
import yaml
import os
import uuid
from jinja2 import Template


class BaseModel(object):

    def __init__(self, model_utility):
        self.model_utility = model_utility

    @staticmethod
    def get_model_id(model_path, rtn_val=False):
        catalog = {}
        index = 0
        model_ids = "Use one of the following models\n"

        for model in os.listdir(model_path):
            if os.path.exists(model_path + model + "/model.json"):
                with open(model_path + model + "/model.json", 'r') as f:
                    model_definition = json.load(f)
                    catalog[index] = model_definition
                    index += 1

        for key in catalog:
            model_ids += "{1}: {0}\n".format(catalog[key]["id"], catalog[key]["name"])

        if rtn_val:
            return catalog

        raise ValueError(model_ids)

    def __get_model_varargs(self, model_id):
        return {
            "model_id": model_id,
            "model_version": "cli",
            "model_table": "AOA_MODELS_cli",
            "project_id": self.__get_project_id(),
            "job_id": str(uuid.uuid4())
        }

    def __get_project_id(self):
        path = os.path.join(self.model_utility.base_path, ".aoa/config.yaml")
        with open(path, "r") as handle:
            return yaml.safe_load(handle)["project_id"]

    def __template_sql_script(self, filename, jinja_ctx):
        with open(filename) as f:
            template = Template(f.read())

        return template.render(jinja_ctx)

    def __execute_sql_script(self, conn, filename, jinja_ctx):
        script = self.__template_sql_script(filename, jinja_ctx)

        stms = script.split(';')

        for stm in stms:
            stm = stm.strip()
            if stm:
                logging.info("Executing statement: {}".format(stm))

                try:
                    conn.execute(stm)
                except Exception as e:
                    if stm.startswith("DROP"):
                        logging.warning("Ignoring DROP statement exception")
                    else:
                        raise e

    def __configure_pyspark(self, model_definition, model_artefacts_path, model_dir, mode):
        import findspark
        import os
        from zipfile import ZipFile
        from os.path import basename, relpath
        import tempfile

        # zip up model files for py-files
        base_zip_path = model_dir + "/model_modules"
        with tempfile.NamedTemporaryFile(delete=False, suffix=".zip") as model_modules_fp:
            with ZipFile(model_modules_fp, 'w') as zip:
                for folder_name, _, filenames in os.walk(base_zip_path):
                    for filename in filenames:
                        full_file_path = os.path.join(folder_name, filename)
                        zip_path = relpath(full_file_path, base_zip_path)
                        zip.write(full_file_path, zip_path)

        # read the spark submit resources for the model
        try:
            # current structure
            if "automation" in model_definition and mode in model_definition["automation"]:
                resources = model_definition["automation"][mode]["resources"]
            # legacy 1
            elif "resources" in model_definition:
                resources = model_definition["resources"]["training"]
            # legacy 2
            else:
                resources = model_definition["automation"]["resources"]["training"]
        except KeyError as e:
            raise Exception("pyspark models require an (automation)->resources->training section in model.json")

        aoa_spark_conf = os.environ.get("AOA_SPARK_CONF", "--conf spark.aoa.modelPath={}".format(model_artefacts_path))

        # legacy support for master being specified as separate property. this should be moved to the args var now
        specific_master_arg = ""
        if "master" in resources:
            specific_master_arg = "--master {}".format(resources["master"])

        os.environ["PYSPARK_SUBMIT_ARGS"] = "{} {} --py-files {} {} pyspark-shell".format(
            specific_master_arg, resources["args"], model_modules_fp.name, aoa_spark_conf)

        # SPARK_HOME should be set
        spark_home = findspark.find()
        findspark.init(spark_home)

        self.logger.info("Using SPARK_HOME: {}".format(spark_home))
        self.logger.info("Using PYSPARK_SUBMIT_ARGS: {}".format(os.environ["PYSPARK_SUBMIT_ARGS"]))
        self.logger.info("Using AOA_SPARK_CONF: {}".format(os.environ.get("AOA_SPARK_CONF", "")))

    def __get_engine(self, model_definition, mode):
        if "automation" in model_definition:

            if mode in model_definition["automation"] and "engine" in model_definition["automation"][mode]:
                return model_definition["automation"][mode]["engine"]

            # legacy
            if "trainingEngine" in model_definition["automation"]:
                return model_definition["automation"]["trainingEngine"]

        return model_definition["language"]

    def __run_r_model(self, model_id, base_path, data_conf, mode):
        import tempfile
        import subprocess

        with tempfile.NamedTemporaryFile(delete=False) as fp:
            fp.write(json.dumps(data_conf).encode())

        cmd = f"{self.model_utility.dir_path}/run_model.R {model_id} {self.__get_project_id()} {mode} {fp.name} {base_path}"
        subprocess.check_call(cmd, shell=True)

